(* Literals. *)
open Core.Std open Looping

module Literal =
  struct
    (* A literal is a triple (sign,proposition name,index).
     * The sign is embedded in the literal because all schemas are considered in negative normal form
     * (i.e. negation sign only in front of atoms).
     *)
    type sign = Positive | Negative
    type t    = sign * Proposition.t * Indexes.t

    (* A literal is loopable, substituable and... is a literal *)
    type substituable = t
    type term         = Indexes.t
    type variable     = Variable.t
    type literal      = t

    (* Constructors *)
    let pos prop idx = (Positive,prop,idx)
    let neg prop idx = (Negative,prop,idx)

    (* Selectors *)
    let is_pos          = function (Positive,_,_) -> true | _ -> false
    let is_neg          = function (Negative,_,_) -> true | _ -> false
    let prop_of l       = Tuple3.get2 l
    let idx_of  l       = Tuple3.get3 l
    let atom_of (_,p,i) = (p,i)

    (* Computes the opposite of a literal *)
    let opposite_sgn = function Positive -> Negative | Negative -> Positive
    let opposite (sgn,prop,idx) = (opposite_sgn sgn,prop,idx)

    (* Print stuff *)
    let sign_to_string = function Positive -> "" | Negative -> "~"
    let to_string (sgn,p,t)              = (sign_to_string sgn) ^ (Proposition.to_string p) ^ "_" ^ (Indexes.to_string t)

    (* Substitution of indexes *)
    let substitute (sgn,p,t') ~var ~by:t = (sgn,p,Indexes.substitute t' ~var ~by:t)
    let compute_lit (sgn,p,t) n = (sgn,p,Indexes.of_num (Indexes.compute_term t n))

    (* Comparisons *)
    let sign_equal sgn1 sgn2 =
      match sgn1,sgn2 with
      |Positive,Positive |Negative,Negative -> true
      |_                                    -> false
    let compare (s1,p1,t1) (s2,p2,t2) =
      match s1,s2 with
      |Positive,Negative -> 1
      |Negative,Positive -> -1
      |_                 -> let pcmp = Proposition.compare p1 p2 in if pcmp <> 0 then pcmp else Indexes.total_compare t1 t2
    let equal (s1,p1,t1) (s2,p2,t2) =
      (match s1,s2 with Positive,Positive | Negative,Negative -> true | _ -> false) && Proposition.equal p1 p2 && Indexes.equal t1 t2

    (* Substitution of propositional variables *)
    module PropSubs : Substitution.Substituable with type substituable = t and type term = Proposition.t and type variable = Proposition.t =
      struct
        type substituable = t
        type term         = Proposition.t
        type variable     = Proposition.t
        let substitute (sgn,p,t) ~var:p' ~by:p'' = if p=p' then (sgn,p'',t) else (sgn,p,t)
      end

    (* Normalization to ease detection of closure *)
    let normalize (s,p,t) ~cstr = (s,p,Constraint.normalize_index cstr ~idx:t)
  end

include Literal

(* All the following for a map on literals. *)

module OrderedSgnProp =
  struct
    type t = sign * Proposition.t
    let compare (s1,p1) (s2,p2) =
      match s1,s2 with
      |Positive,Negative -> 1
      |Negative,Positive -> -1
      |_                 -> Proposition.compare p1 p2
  end
module SgnPropMap = 
  struct
    include Caml.Map.Make(OrderedSgnProp)
    let add m ~key ~data = add key data m
    let find m ~key = try Some (find key m) with Not_found -> None
  end
module LoopingMap =
  struct
    module IdxMap = Indexes.LoopingMap
    type loopable = Literal.t
    type 'a t = ('a IdxMap.t) SgnPropMap.t
    type param = unit
    let empty = SgnPropMap.empty
    let add m ~key:(sgn,prop,idx) ~data = 
      let key = (sgn,prop) in
      SgnPropMap.add ~key ~data:(IdxMap.add (Option.value (SgnPropMap.find ~key m) ~default:IdxMap.empty) ~key:idx ~data) m
    let find m ~key:(sgn,prop,idx) = 
      match SgnPropMap.find ~key:(sgn,prop) m with
      |None    -> None
      |Some m' -> IdxMap.find m' ~key:idx
  end


(* A set of literals. To detect closure.
 *
 * See constraint.ml for more details about the normal form that is used.
 *)
module Map = Caml.Map.Make(struct type t = Literal.t let compare x y = Literal.compare x y end)
module ClosureSet =
  struct
    type t = unit Map.t * Constraint.t
    let add (ls,cstr) ~key = Map.add (Literal.normalize key ~cstr) () ls, cstr
    let closure_applies (ls,cstr) ~key = Map.mem (Literal.normalize (Literal.opposite key) ~cstr) ls
    let empty ~cstr = Map.empty,cstr
    let remove (ls,cstr) ~key = Map.remove (Literal.normalize key ~cstr) ls, cstr
    let add_constraint (ls,cstr) ~cstr:cstr' =
      let cstr'' = Constraint.bounds_conjunction cstr cstr' in
      if Constraint.is_equality cstr then (ls,cstr'') else Map.fold (fun key _ acc -> add acc ~key) ls (empty cstr'')
  end
