%{
  open Core.Std
  open Schema
  open Indexes
  open Constraint

  type param = PropParam of Proposition.t | IndexParam of Variable.t
  type arg   = PropArg of Proposition.t   | IndexArg of Indexes.t
  let add_def (name,params) content k defs = k (Map.add defs ~key:name ~data:(content defs, params))
  let eval_def name args defs = 
    let (content,params) = Map.find_exn defs name in
    try
      List.fold_left2 params args ~init:content 
        ~f:(fun sch param arg ->
          match param,arg with
          |PropParam(pp),PropArg(pa)   -> Schema.PropSubs.substitute sch ~var:pp ~by:pa
          |IndexParam(ip),IndexArg(ia) -> Schema.substitute sch ~var:ip ~by:ia
          |PropParam(_),IndexArg(ia)   ->
              raise (Invalid_argument ("\"" ^ (Indexes.to_string ia) ^ "\" is an index expression whereas a proposition name was expected."))
          |IndexParam(_),PropArg(pa)   ->
              raise (Invalid_argument ("\"" ^ (Proposition.to_string pa) ^ "\" is a proposition name whereas an index expression was expected.")))
    with _ -> raise (Invalid_argument ("Bad number of arguments."))
    
%}
%token <int> INT
%token <string> VAR
%token <string> PROP
%token PLUS MINUS LPAREN RPAREN EQUAL
%token LE LT GE GT
%token AND OR NOT ITAND ITOR XOR EQUIV IMPLY
%token COMMA DOTS ST
%token INDEX
%token DEF LET EOF IN

%left EQUIV IMPLY
%left XOR
%left OR
%left AND
%nonassoc ITAND ITOR 
%nonassoc DOTS EQUAL
%left PLUS MINUS
%nonassoc NOT
%nonassoc LPAREN RPAREN
%nonassoc LET DEF COMMA EOF IN

%start main
%type <Schema.t * Constraint.t> main
%%
main:
  sentences { $1 Map.empty }

sentences:
  LET definition DEF flatschema IN sentences { add_def $2 $4 $6                      } 
  |flatschema EOF                            { fun defs -> ($1 defs, Constraint.top) } 
  |flatschema ST cstr EOF                    { fun defs -> ($1 defs, $3)             } 
;

definition:
  PROP LPAREN params RPAREN { $1, $3 }
  |PROP                     { $1, [] }
;

params:
  PROP               { [PropParam(Proposition.of_string $1)]   } 
  |VAR               { [IndexParam(Variable.of_string $1)]     } 
  |PROP COMMA params { PropParam(Proposition.of_string $1)::$3 } 
  |VAR COMMA params  { IndexParam(Variable.of_string $1)::$3   } 
;

flatschema:
  PROP INDEX terme                                     { fun defs -> pos_lit (Proposition.of_string $1) $3                            } 
  |NOT PROP INDEX terme                                { fun defs -> neg_lit (Proposition.of_string $2) $4                            } 
  |flatschema AND flatschema                           { fun defs -> conj ($1 defs) ($3 defs)                                         } 
  |flatschema OR  flatschema                           { fun defs -> disj ($1 defs) ($3 defs)                                         } 
  |flatschema XOR flatschema                           { fun defs -> xor   ($1 defs) ($3 defs)                                        } 
  |flatschema EQUIV flatschema                         { fun defs -> equiv ($1 defs) ($3 defs)                                        } 
  |flatschema IMPLY  flatschema                        { fun defs -> imply ($1 defs) ($3 defs)                                        } 
  |LPAREN flatschema RPAREN                            { $2                                                                           } 
  |AND VAR EQUAL INT DOTS terme propschema %prec ITAND { fun defs -> it_conj (Variable.of_string $2) (Indexes.of_num $4) $6 ($7 defs) } 
  |OR  VAR EQUAL INT DOTS terme propschema %prec ITOR  { fun defs -> it_disj (Variable.of_string $2) (Indexes.of_num $4) $6 ($7 defs) } 
  |function_call                                       { $1                                                                           } 
;

function_call:
  PROP LPAREN arguments RPAREN { eval_def $1 $3 } 
  |PROP LPAREN RPAREN          { eval_def $1 [] } 
;

arguments:
  terme                  { [IndexArg($1)]                        } 
  |PROP                  { [PropArg(Proposition.of_string $1)]   } 
  |PROP COMMA arguments  { PropArg(Proposition.of_string $1)::$3 } 
  |terme COMMA arguments { IndexArg($1)::$3                      } 
;

propschema:
   PROP INDEX terme              { fun defs -> pos_lit (Proposition.of_string $1) $3 } 
   |NOT PROP INDEX terme         { fun defs -> neg_lit (Proposition.of_string $2) $4 } 
   |propschema AND propschema    { fun defs -> conj  ($1 defs) ($3 defs)             } 
   |propschema OR  propschema    { fun defs -> disj  ($1 defs) ($3 defs)             } 
   |propschema XOR propschema    { fun defs -> xor   ($1 defs) ($3 defs)             } 
   |propschema EQUIV propschema  { fun defs -> equiv ($1 defs) ($3 defs)             } 
   |propschema IMPLY  propschema { fun defs -> imply ($1 defs) ($3 defs)             } 
   |LPAREN propschema RPAREN     { $2                                                } 
   |function_call                                       { $1                                                                           } 
;

cstr:
  VAR LE INT     { constraint_atom (Indexes.of_var (Variable.of_string $1)) Le (Indexes.of_num $3)     } 
  |VAR GE INT    { constraint_atom (Indexes.of_var (Variable.of_string $1)) Gt (Indexes.of_num ($3-1)) } 
  |VAR GT INT    { constraint_atom (Indexes.of_var (Variable.of_string $1)) Gt (Indexes.of_num ($3))   } 
  |VAR LT INT    { constraint_atom (Indexes.of_var (Variable.of_string $1)) Le (Indexes.of_num ($3-1)) } 
  |VAR EQUAL INT { constraint_atom (Indexes.of_var (Variable.of_string $1)) Eq (Indexes.of_num ($3-1)) } 
;

terme:
   VAR            { Indexes.of_var (Variable.of_string $1)             } 
   |INT           { Indexes.of_num $1                                  } 
   |VAR PLUS  INT { add (Indexes.of_var (Variable.of_string $1)) $3    } 
   |VAR MINUS INT { add (Indexes.of_var (Variable.of_string $1)) (-$3) } 
;
