(* Pure literal interface.
 * See developper doc first.
 *)
open Core.Std open Constraint

(* A purity maps pairs ``sign of literal, proposition name of literal'' to sets of indexes.
 * The idea is that if we want to know whether P_e is pure or not,
 * we first get the set of indexes S corresponding to the pair (~,P) and then check if e is in S.
 * In such a case P_e is not pure else it is.
 *)

(* Sets of indexes are represented as unions of intervals and singletons. *)
type domain_unit =
  |Interval  of Indexes.t * Indexes.t
  |Singleton of Indexes.t
type sign = Positive | Negative

module OrderedSgnProp =
  struct
    type t = sign * Proposition.t
    let compare (s1,p1) (s2,p2) =
      match s1,s2 with
      |Positive,Negative -> 1
      |Negative,Positive -> -1
      |_                 -> Proposition.compare p1 p2
  end
module M = Caml.Map.Make(OrderedSgnProp)

(* The type of purities. *)
type t = (domain_unit list) M.t

(* Merging two purities consists in merging the maps and merging the sets of indexes in case of collision of the keys *)
let merge p1 p2 =
  let keys t = M.fold (fun key _ l -> key::l) t [] in
  let all_keys = List.dedup (List.append (keys p1) (keys p2)) in
  List.fold_left ~init:M.empty all_keys ~f:(fun t key ->
    try let data1 = M.find key p1 in
      (try let data2 = M.find key p2 in
        M.add key (data1 @ data2) t
      with Not_found -> M.add key data1 t)
    with Not_found -> M.add key (M.find key p2) t)

(* A range indicates an interval in which a variable occur. *)
type range = Variable.t * Indexes.t * Indexes.t

(* Adding a literal to a purity.
 * - if no range, we just add a singleton with the index
 * - if there is a range, say 1<=i<=n, and the literal contains i, say P_i+1, then we add the interval [2;n+1]
 *)
let add_or_create acc ~key ~data ~combine =
  M.add key (try combine data (M.find key acc) with Not_found -> data) acc
let add ?(range=None) ~lit acc =
  let idx = Literal.idx_of lit in
  add_or_create acc ~key:((if Literal.is_pos lit then Positive else Negative), Literal.prop_of lit) ~combine:(fun l1 l2 -> l1 @ l2) 
  ~data:(match range with
    |None            -> [Singleton idx]
    |Some(var,t1,t2) -> [Interval(Indexes.substitute ~var ~by:t1 idx, Indexes.substitute ~var ~by:t2 idx)])

(* Check if an index is in some set 'dom'.
 * We need the constraint information because e.g. 1 does not occur in [1;n] in general (n may be lower than 1)
 * But if we know that n>=1 then we can ensure that 1 occurs in [1;n]
 *)
let never_in_domain ?(cstr=Constraint.top) ~dom t =
  List.fold_right dom ~init:true
    ~f:(function
      |Singleton(t')   -> (&&) (Constraint.always_different ~hyp:cstr t t')
      |Interval(t1,t2) -> (&&) ((always_greater ~hyp:cstr t1 t) || (always_greater ~hyp:cstr t t2))
    )

(* Says if a literal is pure w.r.t. a purity. *)
let is_pure ~cstr ~pur lit =
  try
    let dom = M.find ((if Literal.is_pos lit then Negative else Positive), (Literal.prop_of lit)) pur in
    never_in_domain ~cstr ~dom (Literal.idx_of lit)
  with Not_found -> true

(* Empty purity = Empty map *)
let empty = M.empty

(* A 'purable' type must provide both a way to update a purity depending on its contents,
 * and a way to remove the pure literals from its contents.
 *)
module type Purable =
  sig
    type purable
    val purity_add : purable -> ?range:range option -> pur:t -> t
    val purify     : purable -> cstr:Constraint.t -> pur:t -> Modified.t * purable
  end
